
/*  @(#)othello.c 1.7 90/04/10
 *
 *  Main routine for the othello program.
 *
 *  Original SunView version by Ed Falk - Sun Microsystems Inc.
 *
 *  Rewritten for independent use by
 *          Rich Burridge, Sun Microsystems, Australia
 *
 *  Copyright (c) Ed Falk & Rich Burridge - Sun Microsystems.
 *                                          All rights reserved.
 *
 *  Permission is given to distribute these sources, as long as the
 /  introductory messages are not removed, and no monies are exchanged.
 *
 *  No responsibility is taken for any errors on inaccuracies inherent
 *  either to the comments or the code of this program, but if reported
 *  to me, then an attempt will be made to fix them.
 */

#include "othello.h"

/* Constants, typedefs, externals, globals, statics, macros. */

char *w_who[] = { "You", "Computer", "White" } ;
char *w_be[]  = { "are", "is", "is" } ;
char *b_who[] = { "Computer", "You", "Black" } ;
char *b_be[]  = { "is", "are", "is" } ;

/* Text values for the cyclic buttons. */
char *asp_values[]  = { "1", "2", "3", "4", "5", "6", NULL } ;
char *diff_values[] = { "1", "2", "3", "4", NULL } ;
char *comp_values[] = { "black", "white", NULL } ;
char *rem_values[]  = { "no", "yes", NULL } ;

enum cantype cmode, last_cmode ;
enum playtype play_mode ;
enum set_type direction ;   /* Incremental direction for cycle item. */

int piece_x ;           /* Current X position of moving piece. */
int piece_y ;           /* Current Y position of moving piece */

int aspire ;            /* Computers level of aspiration. */
int bfont_height ;      /* Height in pixels for bold font. */
int but_inverted ;      /* Value of panel item inverted. */
int color ;             /* Current color value. */
int cretin_flag ;
int cur_ch ;            /* Current character pressed. */
int depth ;             /* Depth of search for computers move. */
int down ;              /* Indicates is a mouse button is down. */
int expected ;
int itemno ;            /* Current panel item being processed. */
int item_value ;        /* Value for current panel item. */
int last_remark ;
int margin ;
int max_bad = 0 ;
int max_good = 0 ;
int move ;              /* Current move being evaluated. */
int move_delta ;        /* Delta for piece animation. */
int nextc ;             /* Current event identifier. */
int next_player ;       /* Next player (BLACK or WHITE) to move. */
int nfont_height ;      /* Height in pixels for normal font. */
int posspec ;           /* Set if -Wp or -g option is present (for X11) */
int remark ;            /* Indicates if remarks are to be displayed. */
int suggestion = -1 ;   /* Positive if a suggested move. */
int suggest_x ;         /* X position of suggested move. */
int suggest_y ;         /* Y position of suggested move. */

/*  Globals for passing arguments to "sandwich";
 *  this is to save time putting arguments on and off the
 *  stack in a very heavily used piece of code
 */

int s_flip ;
int s_move ;
int s_opponent ;
int s_player ;
int s_row ;
int s_col ;

long ab_count ;
long bad_remarks[MAX_REMARKS] ;
long good_remarks[MAX_REMARKS] ;
long se_count ;

char img=0;

FILE *rem_fp ;

struct timeval tp ;        /* Used by the nap_upto routine. */

BOARD old_board ;          /* The previous othello board. */
BOARD board ;              /* The othello board. */
BOARD *s_pos ;

char rem_file[MAXLINE] ;       /* Location of remarks file. */
char bdstring[65];

main(argc,argv)
int argc ;
char *argv[] ;

{
  int taken;

  initialise() ;                /* Initialise variables used by Othello. */
  if (getboard() == -1) {
    printf("Content-type: text/html\n\n");
    printf("<HTML>\n");
    printf("<HEAD>\n");
    printf("<TITLE>Othello - Bad board Error</TITLE>\n");
    printf("</HEAD>\n");
    printf("<BODY>\n");
    printf("<H1>Othello - Bad board error</H1>\n");
    printf("ERROR!  This board is not correct!\n");
    printf("</BODY>\n");
    printf("</HTML>\n");
    exit(0);
  }
  /* print start of board */
  printf("Content-type: text/html\n\n");
  printf("<HTML>\n");
  printf("<HEAD>\n");
  printf("<TITLE>Othello - %d moves left</TITLE>\n",board.moves_left);
  printf("<META HTTP-EQUIV=\"Keywords\" NAME=\"Keywords\" CONTENT=\"Othello,reversi,thomas a. fine\">\n");
  printf("<META NAME=\"robots\" CONTENT=\"index,nofollow\">\n");
  printf("</HEAD>\n");
  printf("<BODY>\n");
  /* for images, leave out this heading to save space */
  if (!img) printf("<H1>Othello - %d moves left</H1>\n",board.moves_left);

  /* handle human move */
  move=getmove();
  switch(move) {
    case -1:
      printf("OTHELLO ERROR!  That wasn't a valid move!\n");
      printf("</BODY>\n");
      printf("</HTML>\n");
      exit(0);
      break;
    case START:
      printf("Welcome to Othello.<P>\n");
      if (!img) {
	printf("You can also try the\n");
	printf("<A NAME=img HREF=\"http://hea-www.harvard.edu/~fine/cgi/othello.cgi/img/\">\n");
	printf("inlined image</A> mode, which is known to\n");
	printf("work with at least Mosaic for X 2.4.\n");
	printf("[you can tell I wrote this ages ago - it should work with any modern browser --tom]<P>\n");
      }
      printf("Please select a move:\n");
      draw_board(WHITE);
      printf("Or you may choose to <A HREF=\"http://hea-www.harvard.edu/~fine/cgi/othello.cgi/%s%s?pass\">pass</A>.\n", img?"img/":"", bdstring);
      showrules();
      printf("</BODY>\n");
      printf("</HTML>\n");
      exit(0);
      break;
    case PASS:
      printf("<P>\nYou pass, my move:\n");
      break;
    default:
      if (legal(move, BLACK, &board) == FALSE) {
	printf("OTHELLO ERROR!  That wasn't a legal move!\n");
	printf("</BODY>\n");
	printf("</HTML>\n");
	exit(0);
      } else {
	taken=formfliplist(move, BLACK) ;
	printf("<P>\nYou took %d pieces\n",taken);
	draw_board(BLACK);
      }
      break;
  }
  /* did you just win? */
  if (game_over()) {
    printf("</BODY>\n");
    printf("</HTML>\n");
    exit(0);
  }

  /* do computer move */
  move = makemove(&board, WHITE, depth) ;
  if (move == NOMOVE) {
    printf("<P>\nComputer passes, select a move:\n");
    if (draw_board(WHITE) == 0) {
      printf("EEE GADS!  A stalemate.  Game over.\n");
    } else {
      printf("Or you may choose to <A HREF=\"http://hea-www.harvard.edu/~fine/cgi/othello.cgi/%s%s?pass\">pass</A>.\n", img?"img/":"", bdstring);
    }
  } else {
    taken=formfliplist(move, WHITE) ;
    printf("<P>\nComputer took %d pieces\n",taken);
    if (draw_board(WHITE) == 0) {
      if (game_over()) {
	printf("</BODY>\n");
	printf("</HTML>\n");
	exit(0);
      } else {
	printf("You have no moves.  You must <A HREF=\"http://hea-www.harvard.edu/~fine/cgi/othello.cgi/%s%s?pass\">pass</A>.\n", img?"img/":"", bdstring);
      }
    } else {
      printf("Or you may choose to <A HREF=\"http://hea-www.harvard.edu/~fine/cgi/othello.cgi/%s%s?pass\">pass</A>.\n", img?"img/":"", bdstring);
    }
  }

  /* get_remarks() ; */
  printf("</BODY>\n");
  printf("</HTML>\n");
  exit(0);
}

int game_over()
{
  int cs, ps ;

  ps = count(&board, WHITE) ;
  cs = count(&board, BLACK) ;
  if ((!cs) || (!ps) || ((cs+ps) == 64)) {
    if (ps > cs) {
      printf("The Computer wins %d-%d\n", ps, cs) ;
    } else if (ps == cs) {
      printf("A tie %d-%d\n", ps, cs) ;
    } else {
      printf("You win %d-%d\n", cs, ps) ;
    }
    return(1);
  }
  return(0);
}

int draw_board(justwent)
int justwent;
{
  int i, j, nummoves=0;

  if (img) return(img_draw_board(justwent));

  for (j=0; j<8; ++j) {
    for (i=0; i<8; ++i) {
      if (board.square[j*8+i] == FREE) bdstring[j*8+i]='-';
      else if (board.square[j*8+i] == BLACK) bdstring[j*8+i]='x';
      else if (board.square[j*8+i] == WHITE) bdstring[j*8+i]='o';
    }
  }
  bdstring[64]='\0';

  printf("<PRE>\n");
  for (j=0; j<8; ++j) {
    printf("  +---+---+---+---+---+---+---+---+\n  ");
    for (i=0; i<8; ++i) {
      if (board.square[j*8+i] == FREE) {
	if (justwent==WHITE) {
	  if (legal(j*8+i,BLACK,&board)) {
	    printf("|<A HREF=\"http://hea-www.harvard.edu/~fine/cgi/othello.cgi/%s?%d,%d\">   </A>",bdstring,i,j);
	    ++nummoves;
	  }  else {
	    printf("|   ");
	  }
	} else {
	  printf("|   ");
	}
      } else if (board.square[j*8+i] == WHITE) {
	if (old_board.square[j*8+i] == WHITE) {
	  printf("| o ");
	} else if (old_board.square[j*8+i] == BLACK) {
	  printf("| O ");
	} else {
	  printf("| <strong>O</strong> ");
	}
      } else if (board.square[j*8+i] == BLACK) {
	if (old_board.square[j*8+i] == BLACK) {
	  printf("| x ");
	} else if (old_board.square[j*8+i] == WHITE) {
	  printf("| X ");
	} else {
	  printf("| <strong>X</strong> ");
	}
      }
    }
    printf("|\n");
  }
  printf("  +---+---+---+---+---+---+---+---+\n");
  printf("</PRE>\n");
  return(nummoves);
}

int img_draw_board(justwent)
int justwent;
{
  int i, j, nummoves=0;

  for (j=0; j<8; ++j) {
    for (i=0; i<8; ++i) {
      if (board.square[j*8+i] == FREE) bdstring[j*8+i]='-';
      else if (board.square[j*8+i] == BLACK) bdstring[j*8+i]='x';
      else if (board.square[j*8+i] == WHITE) bdstring[j*8+i]='o';
    }
  }
  bdstring[64]='\0';

  printf("<P>\n");
  for (j=0; j<8; ++j) {
    for (i=0; i<8; ++i) {
      if (board.square[j*8+i] == FREE) {
	if (justwent==WHITE) {
	  if (legal(j*8+i,BLACK,&board)) {
	    printf("<A HREF=\"%s?%d,%d\"><IMG SRC=\"0\"></A>",bdstring,i,j);
	    ++nummoves;
	  }  else {
	    printf("<A HREF=\"\"><IMG SRC=\"0\"></A>");
	  }
	} else {
	  printf("<A HREF=\"\"><IMG SRC=\"0\"></A>");
	}
      } else if (board.square[j*8+i] == WHITE) {
	if (old_board.square[j*8+i] == WHITE) {
	  printf("<A HREF=\"\"><IMG SRC=\"w\"></A>");
	} else if (old_board.square[j*8+i] == BLACK) {
	  printf("<A HREF=\"\"><IMG SRC=\"x\"></A>");
	} else {
	  printf("<A HREF=\"\"><IMG SRC=\"y\"></A>");
	}
      } else if (board.square[j*8+i] == BLACK) {
	if (old_board.square[j*8+i] == BLACK) {
	  printf("<A HREF=\"\"><IMG SRC=\"b\"></A>");
	} else if (old_board.square[j*8+i] == WHITE) {
	  printf("<A HREF=\"\"><IMG SRC=\"c\"></A>");
	} else {
	  printf("<A HREF=\"\"><IMG SRC=\"d\"></A>");
	}
      }
    }
    printf("<br>\n");
  }
  return(nummoves);
}

int getboard()
{
  int i, j, num=64;
  char *boardstring;

  boardstring=getenv("PATH_INFO");
  if (boardstring == NULL) boardstring="";
  if (strncmp(boardstring,"/img/",5) == 0) {
    img=1;
    boardstring += 5;
  }
  while (*boardstring == '/') ++boardstring;

  if (img && (strlen(boardstring)==1)) {
    return_gif(boardstring);
  }

  if (strlen(boardstring)==0) {
    FOR_BOARD(i) old_board.square[i] = board.square[i] = FREE ;
    old_board.square[27] = board.square[27] = WHITE ;
    old_board.square[28] = board.square[28] = BLACK ;
    old_board.square[35] = board.square[35] = BLACK ;
    old_board.square[36] = board.square[36] = WHITE ;
    board.moves_left = 60 ;
    num=60;
    return(num);
  }

  for (j=0; j<8; ++j) {
    for (i=0; i<8; ++i) {
      switch(*boardstring) {
	case '-':
	  board.square[j*8+i] = FREE;
	  old_board.square[j*8+i] = FREE;
	  break;
	case 'o':
	  board.square[j*8+i] = WHITE;
	  old_board.square[j*8+i] = WHITE;
	  --num;
	  break;
	case 'x':
	  board.square[j*8+i] = BLACK;
	  old_board.square[j*8+i] = BLACK;
	  --num;
	  break;
	case '\0':
	  return(-1);
	  break;
	default:
	  return(-1);
	  break;
      }
      ++boardstring;
    }
  }
  board.moves_left=num;
  old_board.moves_left=num;
  return(num);
}

return_gif(c)
char *c;
{

  switch(*c) {
    case '0':
      printf("Location: /~fine/images/blank.gif\n\n");
      break;
    case 'w':
      printf("Location: /~fine/images/white.gif\n\n");
      break;
    case 'x':
      printf("Location: /~fine/images/white-flip.gif\n\n");
      break;
    case 'y':
      printf("Location: /~fine/images/white-new.gif\n\n");
      break;
    case 'b':
      printf("Location: /~fine/images/black.gif\n\n");
      break;
    case 'c':
      printf("Location: /~fine/images/black-flip.gif\n\n");
      break;
    case 'd':
      printf("Location: /~fine/images/black-new.gif\n\n");
      break;
    default:
      printf("Content-type: text/html\n\n");
      printf("<HTML>\n");
      printf("<HEAD>\n");
      printf("<TITLE>Othello - Bad board Error</TITLE>\n");
      printf("</HEAD>\n");
      printf("<BODY>\n");
      printf("<H1>Othello - Bad board error</H1>\n");
      printf("ERROR!  This image specification is not correct!\n");
      printf("</BODY>\n");
      printf("</HTML>\n");
      break;
  }
  exit(0);
}

getmove()
{
  int i, x, y, ret;
  char *movestring, *xstr, *ystr;

  movestring=getenv("QUERY_STRING");
  if (movestring == NULL) movestring="";
  if (strcmp(movestring,"pass")==0) return(PASS);
  if (strlen(movestring)==0) return(START);
  xstr=movestring;
  for (i=0; i<strlen(movestring); ++i) {
    if (movestring[i] == ',') {
      ystr=movestring+i+1;
      movestring[i]='\0';
      break;
    }
  }
  x=atol(xstr);
  y=atol(ystr);
  if (x<0 || x>7) return(-1);
  if (y<0 || y>7) return(-1);
  ret=y*8+x;
  return(ret);
}

initialise()        /* Initialise various variable used by Othello. */
{
  /* remark = items[(int) REMARK].value ; */
  play_mode = PLAY_WHITE ;
  cmode = BLACK_START ;
  aspire = INIT_ASPIRE ;
  depth = INIT_DEPTH ;
  cretin_flag = FALSE ;
  last_remark = FIRST_REMARK ;        /* Force a remark the first time. */
  STRCPY(rem_file, REMNAME) ;
}

count(board, player)    /* Count the number of player pieces on the board. */
int player ;
BOARD *board ;
{
  int i, n ;

  n = 0 ;
  FOR_BOARD(i)
    if (board->square[i] == player) n++ ;
  return(n) ;
}

formfliplist(move, player)
int move, player ;
{
  int cnt, i, old_cnt ;

  old_cnt = count(&board, player) ;
  FOR_BOARD(i) old_board.square[i] = board.square[i] ;
  old_board.moves_left = board.moves_left ;
  domove(&old_board, move, &board, player) ;
  cnt = count(&board, player) ;
  return(cnt - old_cnt - 1) ;
}

showrules()
{
  char *info1="\
<H1>About othello</H1>\n\
This version of Othello (which is probably somebodies trademark) uses\n\
the strategy of Chris Miller, which was ripped from the Othello 1.3\n\
developed by Rich Burridge and Ed Falk.  The game was converted to\n\
work with the web (by throwing away large quantities of code) by\n\
<A NAME=taf HREF=\"http://hea-www.harvard.edu/~fine/\">Thomas A. Fine</A>\n\
in early 1994 (inlined images were still new, hence text is the default).\n";

  char *info2a="\
<H1>Using The Web Interface</H1>\n\
You are currently using the ascii interface.  <TT>X</TT> is black (which you\n\
will always play), and <TT>O</TT> is white (always the computer).  If\n\
you want to go second, you can always pass on the first move.\n\
<P>\n\
Your options for moves will be blank squares marked as hypertext links.\n\
Click on a link to put a piece there (yes, this makes it easier).\n\
If there are no links showing, either the\n\
game is over (it will say so), or you have to pass (which it will also\n\
say).\n\
<P>\n\
When the board is redrawn to show the results of a move (yours, the\n\
computers, or both) all unchanged pieces are shown in lower case.\n\
Any pieces which changed will be upper case, including the new piece\n\
which will also be boldface.\n";

  char *info2i="\
<H1>Using The Web Interface</H1>\n\
You are currently using the graphic interface.  You will always\n\
play black, and will go first normally.  If\n\
you want to go second, you can always pass on the first move.\n\
<P>\n\
To get the proper formatting with inlined images, all squares are hypertext\n\
links.  Most are empty links, which simply causes the page to redisplay\n\
when using Mosaic for X.  I have no idea what happens under other browsers;\n\
in fact I'm not even sure if the board will display in a reasonable fashion.\n\
<P>\n\
Your options for moves will be hypertext links that are not empty.\n\
Mosaic for X will display the link address, so you can tell.  I'm not\n\
sure if other browsers allow this.  At any rate\n\
click on a square to put a piece there.\n\
If there are no moves showing, either the\n\
game is over (it will say so), or you have to pass (which it will also\n\
say).\n\
<P>\n\
When the board is redrawn to show the results of a move (yours, the\n\
computers, or both) all unchanged pieces are shown on a green square.\n\
Any pieces which were flipped will be on yellow squares, and the new piece\n\
will be on a red square.\n";

  char *info3="\
<H1>Basic Rules</H1>\n\
The idea is to capture all of your opponents pieces, or to have more pieces\n\
than your opponent at game's end (when there are no more moves).\n\
<P>\n\
To capture pieces, you move your piece onto a square adjacent to a row,\n\
column, or diagonal of your opponents pieces.  Another one of your\n\
pieces must be found at the other end of that row, column, or\n\
diagonal.  All the pieces between your original piece and your new\n\
piece are then converted into your game color (flipped over in the\n\
physical world).\n\
<P>\n\
If there is more than one row, column, or diagonal that can be influenced\n\
by a new piece being added, they will <em>all</em> be flipped.\n\
<P>\n\
If you can not capture any pieces, you must pass.  You can not place\n\
pieces without capturing.\n\
<P>\n\
Have fun.\n\
";

  puts(info1);
  if (img) puts(info2i);
  else puts(info2a);
  puts(info3);
}
